const PubSub = require('pubsub-js');
const Boom = require('boom');

class SocketServer {
    constructor(){
        this.clientsMap = new Map();
        this.subscriptionsMap = new Map();
        this.subscriber = (action, callback) => {
            PubSub.subscribe(action, (action, data) => callback(data))
        };
        this.publisher = PubSub.publish;
    }

    set({publisher, subscriber}) {
        if (!this.subscriptionsMap.size) {
            if (publisher) this.publisher = publisher;
            if (subscriber) this.subscriber = subscriber;
        }
        else {
            console.error('cannot set pubsub after set subscriptions');
        }
    }

    listen(server, callback = ()=>{}){
        this.io = require('socket.io')(server);
        this.io.on('connection', this._manager.bind(this));
        callback(this.io);
    }

    _manager(socket){
        socket.on('disconnect', () => this._onDisconnect(socket));
        socket.on('SUBSCRIBE', request=>this._onSubscribe(socket, request));
        socket.on('UPDATE_SUBSCRIPTION', request => this._onUpdateSubscription(socket, request));
        socket.on('UNSUBSCRIBE', request => this._onUnsubscribe(socket, request.action));
    }

    _onDisconnect(socket) {
        this.clientsMap.forEach((clients, action) => this._onUnsubscribe(socket, action, 'disconnected', clients));
    }

    _onUpdateSubscription(socket, request) {
        this._onUnsubscribe(socket, request.action, 'update');
        this._onSubscribe(socket, request);
    }

    _onSubscribe(socket, request){
        if(this.subscriptionsMap.has(request.action)){
            let response = {
                error : (err)=>this._onSubscribeError(socket, request, err),
                success : ()=>this._onSubscribeSuccess(socket, request),
            };
            this._middlewaresExec(
                request,
                response,
                this._preMiddleware,
                ...this.subscriptionsMap.get(request.action).middlewares,
                this._postMiddleware,
            );
        }
    }

    _onSubscribeSuccess(socket, request){
        let clientsForAction = this.clientsMap.has(request.action) ? this.clientsMap.get(request.action) : [];
        let newClient = {...request, socketId: socket.id};
        this.clientsMap.set(request.action, [...clientsForAction, newClient]);
        let subscription = this.subscriptionsMap.get(request.action);
        if (typeof subscription.onSubscribe === 'function') {
            subscription.onSubscribe(newClient);
        }
    }

    _onSubscribeError(socket, request, unCleanError){
        let error = (unCleanError.isBoom) ? unCleanError : Boom.badImplementation();
        let {action, subcriptionId, params} = request;
        socket.emit(request.action, {
            action,
            subcriptionId,
            params,
            data : {},
            error : error.output.payload
        });
        let subscription = this.subscriptionsMap.get(request.action);
        if (typeof subscription.onSubscribeError === 'function') {
            subscription.onSubscribeError(error.output.payload);
        }
    }

    _onUnsubscribe(socket, action, reason = 'unsubscribed', clients) {
        if (clients || this.clientsMap.has(action)) {
            clients = clients || this.clientsMap.get(action);
            let clientsWithoutDisconnected = clients.filter(client=>client.socketId !== socket.id);
            (clientsWithoutDisconnected.length >= 1) ? this.clientsMap.set(action, clientsWithoutDisconnected) : this.clientsMap.delete(action);
            let subscription = this.subscriptionsMap.get(action);
            if (typeof subscription.onUnsubscribe === 'function') {
                let client = clients.find(client => client.socketId === socket.id);
                client && subscription.onUnsubscribe(client, reason);
            }
        }
    }

    _preMiddleware(req, res, next){
        req.isSocket = true;
        next();
    }

    _postMiddleware(req, res){
        res.success();
    }

    async _middlewaresExec(request = {}, response = {}, ...funcs){
        try {
            await funcs.reduceRight((composed, f) => async function (err) {
                err ? response.error(err) : await f(request, response, composed);
            }, ()=>{})();
        }
        catch (err) {
            response.error(err);
        }
    }

    _publishSockets(action = '', clients = [], data = {}) {
        clients.forEach(client => this.io.to(client.socketId).emit(action, {
            subcriptionId : client.subcriptionId,
            params : client.params,
            action,
            data
        }));
    }

    _publish(action, data = {}){
        let subscription = this.subscriptionsMap.get(action);
        let clients = this.clientsMap.has(action) ? this.clientsMap.get(action) : [];
        let responseData = (typeof subscription.transformResponse === 'function' && typeof data === 'object') ?  subscription.transformResponse({...data}) : data;
        if (typeof subscription.filter === 'function'){
            try {
                 let filterData = (typeof subscription.preFilter === 'function') ?  subscription.preFilter(data) : data;
                 this._publishSockets(action, clients.filter(client => subscription.filter(client, filterData)), responseData);
            }
            catch (e) {
                this._publishSockets(action, [], responseData); //Oups no error handler :/
            }
        }
        else {
            this._publishSockets(action, clients, responseData);
        }
    }

    publish(action, data) {
        this.publisher(action, data);
    }

    subscription(action, ...middlewares) {
        this.subscriptionsMap.set(action, {middlewares});
        this.subscriber(action, (data) => this._publish(action, data));
        return {
            use: (options) => {
                this.subscriptionsMap.set(action, {middlewares, ...options})
            }
        }
    }
}

module.exports = new SocketServer;